// arithmetic_traits.h -*-c++-*-
#ifndef _arithmetic_traits_HDR_
#define _arithmetic_traits_HDR_

#include<boost/static_assert.hpp>
// #include<boost/type_traits.hpp> //makes problems with "long long" messages
#include<boost/mpl/apply_if.hpp>
#include<boost/mpl/not.hpp>

#include<complex>
// namespace pu{ //workaround
//   template<typename A,typename B> struct is_same{
//     enum {value = false};
//   };
//   template<typename A> struct is_same<A,A>{
//     enum {value = true};
//   };
// }
namespace arith_traits_pr{
  using namespace boost;
  using namespace boost::mpl;
  using namespace pu;
  template<typename SC1,typename SC2> struct helper_arith_promote{
    typedef typename
    apply_if<is_same<SC1,SC2>,identity<SC1>,
	     apply_if<is_same<std::complex<SC1>,SC2 >,identity<SC2>,
		      apply_if<is_same<SC1,std::complex<SC2> >,identity<SC1>,
			       identity<void> > > >::type promote_type;
    BOOST_STATIC_ASSERT((!is_same<promote_type,void>::value));
  };
}//of arith_traits_pr

namespace pu {
  template<typename SC1,typename SC2> struct gen_arith_promotion_traits{
    typedef typename arith_traits_pr::helper_arith_promote<SC1,SC2>::
    promote_type promote_type;
  };

  template<typename SC1,typename SC2> struct add_promotion_traits{
    typedef typename
    gen_arith_promotion_traits<SC1,SC2>::promote_type promote_type;
  };
  template<typename SC1,typename SC2> struct mult_promotion_traits{
    typedef typename
    gen_arith_promotion_traits<SC1,SC2>::promote_type promote_type;
  };
  template<typename SC1,typename SC2> struct div_promotion_traits{
    typedef typename
    gen_arith_promotion_traits<SC1,SC2>::promote_type promote_type;
  };

  template<typename SC,typename Q> class multassign{
    //use only as temporary object
  private:
    Q const &s;
  public:
    multassign(Q const & in):s(in) {}
    void operator()(SC & target) const { target *= s; }
  };
  template<typename SC,typename Q> class divideassign{
    //use only as temporary object
  private:
    Q const &s;
  public:
    divideassign(Q const & in):s(in) {}
    void operator()(SC & target) const { target /= s; }
  };
  template<typename S1,typename S2> struct mixplus{
    typename add_promotion_traits<S1,S2>::promote_type const
    operator()(S1 const & s1,S2 const & s2) const { return s1+s2; }
  };
  template<typename S1,typename S2> struct mixminus{
    typename add_promotion_traits<S1,S2>::promote_type const
    operator()(S1 const & s1,S2 const & s2) const { return s1-s2; }
  };
  template<typename S1,typename S2> struct mixdiv{
    typename div_promotion_traits<S1,S2>::promote_type const
    operator()(S1 const & s1,S2 const & s2) const { return s1/s2; }
  };
  template<typename S1,typename S2> struct mixmult{
    typename mult_promotion_traits<S1,S2>::promote_type const
    operator()(S1 const & s1,S2 const & s2) const { return s1*s2; }
  };


  // for treating complex and real scalars
  template<typename S> struct real_part{ typedef S type; };
  template<typename S> struct real_part<std::complex<S> >{ typedef S type; };
  template<typename SC> class conjugator{
  public: void operator()(SC & target){};
  };
  template<typename SC> class conjugator<std::complex<SC> >{
  public: void operator()(SC & target){target = std::conj(target);};
  };
  template<typename SC> SC const extract_real(SC const & s) { return s; }
  template<typename SC> SC const extract_real(std::complex<SC> const & s) { 
    return s.real() ; 
  }

} //of pu

#endif
