//vnd.h -*-c++-*-
#ifndef _vnd_HDR_
#define _vnd_HDR_

#include"pr1arr.h"
#include"arithmetic_tools.h"
// #include"pr1io.h"

#include<algorithm>
#include<functional>
#include<iterator>
#include<complex>
#include<iostream> // for std::cerr
#include<cstdlib> // for std::exit()
#include<numeric> // for std::inner_product.

#include<boost/static_assert.hpp>
#include<boost/mpl/apply_if.hpp>

namespace vnd_pr {
  typedef unsigned int uinteger;
  template<unsigned int N> uinteger dim(){ return N; }
  template<unsigned int N> void ck_ix(uinteger j) { // helper
    if(( j < 0 ) || ( N <= j )){
      std::cerr << "vnd<SC,"<< N << ">::[j] wrong j = " << j << std::endl ;
      std::exit(-1);
    }
  } 
  template<typename IT>
  std::ostream & print(std::ostream & C, IT const &beg , IT const & end){
    typedef typename std::iterator_traits<IT>::value_type SC;
    C << "(" ;
    std::copy(beg,end-1,std::ostream_iterator<SC>(C,","));
    if (end > beg){
      C << *(end-1) ;
    }
    return C << ")" ;    
  }
  template<typename IT>
  std::istream & get(std::istream & C,IT beg , IT const & end){
    for(;beg != end; ++beg){
      C >> *beg;
    }
    return C;
  }
} // of namespace vnd_pr

namespace pu{
  //-----------------------------------------------//
  // some try of use of MPL in the following class //
  //-----------------------------------------------//
  template<typename SC,unsigned int N> class vnd{
    // assumption: "SC(),Q(),R() etc." should have zero semantics.
  private:
    typedef n_arr<SC,N> inner_arr_t;
    inner_arr_t dat;

    struct nofill_optimization {};
    vnd(nofill_optimization) {} // for intermediate results

  public:
    typedef typename inner_arr_t::iterator iterator;
    typedef typename inner_arr_t::const_iterator const_iterator;

    vnd() :dat(SC()) { //to keep zero semantic
    }

    iterator begin() { return dat.begin(); }
    iterator end() { return dat.end(); }
    const_iterator begin() const { return dat.begin(); }
    const_iterator end() const { return dat.end(); }
    ~vnd() {}

    SC &  operator[](unsigned int const i) { return dat[i]; }
    SC const & operator[](unsigned int const i) const { return dat[i]; }

    std::ostream & print(std::ostream & C) const {
      return vnd_pr::print(C,begin(),end()); }
    std::istream & get(std::istream & C) {
      return vnd_pr::get(C,begin(),end()); }

    template<typename Q>
    vnd & operator*= (Q const & s){
      std::for_each(begin(),end(),multassign<SC,Q>(s));
      return *this;
    }
    template<typename Q>
    vnd & operator /= (Q const & s){
      std::for_each(begin(),end(),divideassign<SC,Q>(s));
      return *this;
    }
    template<typename Q>
    vnd & operator+=(vnd<Q,N> const & v) {
      std::transform(begin(),end(),v.begin(),begin(),mixplus<SC,Q>());
      return *this;
    }
    template<typename Q>
    vnd & operator-=(vnd<Q,N> const & v) {
      std::transform(begin(),end(),v.begin(),begin(),mixminus<SC,Q>());
      return *this;
    }

    // assuming that the set {scalars of type SC} has a ring properities
    vnd const operator+(vnd const & v)const{ return vnd(*this).operator+=(v); }
    vnd const operator-(vnd const & v)const{ return vnd(*this).operator-=(v); }
    vnd const operator*(SC const & s)const{ return vnd(*this).operator*=(s); }
    vnd const operator/(SC const & s)const{ return vnd(*this).operator/=(s); }

    template<typename Q> struct add_res_sc{
      typedef typename add_promotion_traits<SC,Q>::promote_type type;
    };
    template<typename Q> struct mult_res_sc{
      typedef typename mult_promotion_traits<SC,Q>::promote_type type;
    };
    template<typename Q> struct div_res_sc{
      typedef typename div_promotion_traits<SC,Q>::promote_type type;
    };

    template<typename Q>
    vnd<typename add_res_sc<Q>::type,N> const 
    operator+(vnd<Q,N> const &v)const {
      vnd<typename add_res_sc<Q>::type,N> result(nofill_optimization());
      std::transform(begin(),end(),v.begin(),result.begin(),mixplus<SC,Q>());
      return result;
    }
    template<typename Q>
    vnd<typename add_res_sc<Q>::type,N> const 
    operator-(vnd<Q,N> const &v)const {
      vnd<typename add_res_sc<Q>::type,N> result(nofill_optimization());
      std::transform(begin(),end(),v.begin(),result.begin(),mixminus<SC,Q>());
      return result;
    }
    template<typename Q>
    vnd<typename mult_res_sc<Q>::type,N> const 
    operator*(Q const &q)const {
      vnd<typename mult_res_sc<Q>::type,N> res(nofill_optimization());
      std::transform(begin(),end(),res.begin(),
		     std::bind2nd(mixmult<SC,Q>(),q));
      return res;
    }
    template<typename Q>
    vnd<typename div_res_sc<Q>::type,N> const 
    operator/(Q const &q)const {
      vnd<typename div_res_sc<Q>::type,N> res(nofill_optimization());
      std::transform(begin(),end(),res.begin(),std::bind2nd(mixdiv<SC,Q>(),q));
      return res;
    }

    template<typename Q>
    typename mult_res_sc<Q>::type const 
    dot(vnd<Q,N> const &v) const {
      typedef typename mult_res_sc<Q>::type R;
      return std::inner_product(begin(),end(),v.begin(),R(),
				std::plus<R>(),mixmult<SC,Q>());
    }

    typedef typename real_part<SC>::type real_part;
  private:
    vnd const conj_copy() const{
      return vnd(*this).conj();
    }
  public:
    real_part const norm2() const { 
      return std::sqrt(extract_real(conj_copy().dot(*this)));
    }
    vnd const & conj() {
      std::for_each(begin(),end(),conjugator<SC>());
      return *this;
    }

    //only when N==3
    vnd(SC const & x,SC const & y,SC const & z);

    //only when N==2
    vnd(SC const & x,SC const & y);

    template<typename RT,typename T1,typename T2> struct cross3_t{
      static RT const cross(T1 const &v1,T2 const &v2){
	return RT(v1[1]*v2[2] - v2[1]*v1[2],
		  v1[2]*v2[0] - v2[2]*v1[0],
		  v1[0]*v2[1] - v2[0]*v1[1]);
      }
    };
    template<typename RT,typename T1,typename T2> struct cross2_t{
      static RT const cross(T1 const &v1,T2 const &v2){
	return RT(v1[0]*v2[1] - v2[0]*v1[1]);
      }
    };
    template<typename Q> struct cross_res_t{
      typedef typename boost::mpl::
      apply_if< boost::mpl::bool_<(N==2)>,
		boost::mpl::identity<mult_res_sc<Q> >,
		boost::mpl::apply_if<
	boost::mpl::bool_<(N==3)>,
	boost::mpl::identity<vnd<typename mult_res_sc<Q>::type,3> >,
	boost::mpl::identity<void> > >::type
      type;
      BOOST_STATIC_ASSERT(!(boost::is_same<type,void>::value));

      typedef typename boost::mpl::
      apply_if<boost::mpl::bool_<N==3>,
	       boost::mpl::identity<cross3_t<type,vnd<SC,3>,vnd<Q,3> > >,
	       boost::mpl::identity<cross2_t<type,vnd<SC,2>,vnd<Q,2> > > >
      ::type action_t;
    };

    template<typename Q> typename cross_res_t<Q>::type const
    cross(vnd<Q,N> const & v) const ; // returns a copy
  };

  template<typename SC,unsigned int N> template<typename Q>
  typename vnd<SC,N>::template cross_res_t<Q>::type const
  vnd<SC,N>::cross(vnd<Q,N> const & b) const {
    BOOST_STATIC_ASSERT(((N==2)||(N==3)));
    namespace mpl=boost::mpl;
    typedef typename cross_res_t<Q>::type result_type;
    typedef typename cross_res_t<Q>::action_t action_t;
    return action_t::cross(*this,b);
  }

  template<typename SC,unsigned int N>
  vnd<SC,N>::vnd(SC const & x,SC const & y,SC const & z) {
    BOOST_STATIC_ASSERT(N==3);
    // to force the compilation error rather then linking
    dat[0] = x; dat[1] = y; dat[2] = z;
  }

  template<typename SC,unsigned int N>
  vnd<SC,N>::vnd(SC const & x,SC const & y) {
    BOOST_STATIC_ASSERT(N==2);
    dat[0] = x; dat[1] = y;
  }

  // non-members
  template<typename SC,unsigned int N,typename Q>
  typename vnd<SC,N>::template cross_res_t<Q>::type const
  cross(vnd<SC,N> const & a,vnd<Q,N> const & b){ return a.cross(b); }

  template<typename SC,unsigned int N> vnd<SC,N> const
  conj(vnd<SC,N> v){return v.conj();}

  template<typename SC,unsigned int N> std::ostream &
  operator<< (std::ostream & C,vnd<SC,N> const & a){ return a.print(C); }

  template<typename SC,unsigned int N> std::istream & 
  operator >> (std::istream & C,vnd<SC,N> & a){ return a.get(C);}

  template<typename SC,unsigned int N,typename Q>
  typename vnd<SC,N>::template mult_res_sc<Q>::type const
  dot(vnd<SC,N> const &v1,vnd<Q,N> const &v2){return v1.dot(v2); }

  template<typename SC,unsigned int N> typename vnd<SC,N>::real_part const
  norm2(vnd<SC,N> v){return v.norm2(); }

  template<typename Q,typename SC,unsigned int N>
  vnd<typename mult_promotion_traits<Q,SC>::promote_type,N> const
  operator*(Q const &q,vnd<SC,N> const & v)
  {return v*q;}


} //of pu

namespace pu {
  template<class SC> class v3d{
  private:
    vnd<SC,3> data;
    explicit v3d(vnd<SC,3> const &v_):data(v_) {}
  public:
    v3d() {}
    v3d(SC const & x,SC const & y,SC const & z) : data(x,y,z){}
    SC & operator[](unsigned int i) { return data[i]; }
    SC const &  operator[](unsigned int i) const { return data[i]; }
    std::ostream & print(std::ostream & C) const { return data.print(C); }
    std::istream & get(std::istream & C) { return data.get(C); }

    template<typename Q>
    typename vnd<SC,3>::template cross_res_t<Q>::type const
    cross(v3d<Q> const & p){ return 
	typename vnd<SC,3>::template cross_res_t<Q>::type(data.cross(p.data));
    }

    template<typename Q>
    v3d<SC> & operator*=(Q const & s){ data*= s; return *this; }
    template<typename Q>
    v3d<SC> & operator/=(Q const & s){ data/= s; return *this; }
    v3d<SC> operator+=(v3d<SC> const &v){
      data+=v.data; return *this;
    }
    v3d<SC> operator-=(v3d<SC> const &v){
      data-=v.data; return *this;
    }

    template<typename Q>
    v3d<typename vnd<SC,3>::template add_res_sc<Q>::type> const
    operator+(v3d<Q> const &v) const {
      return v3d<typename vnd<SC,3>::template add_res_sc<Q>::type>
	(data+v.data);
    }
    template<typename Q>
    v3d<typename vnd<SC,3>::template add_res_sc<Q>::type> const
    operator-(v3d<Q> const &v) const {
      return v3d<typename vnd<SC,3>::template add_res_sc<Q>::type>
	(data-v.data);
    }
    template<typename Q>
    v3d<typename vnd<SC,3>::template mult_res_sc<Q>::type> const
    operator*(Q const &q) const {
      return v3d<typename vnd<SC,3>::template mult_res_sc<Q>::type>
	(data*q);
    }
    template<typename Q>
    v3d<typename vnd<SC,3>::template div_res_sc<Q>::type> const
    operator/(Q const &q) const {
      return v3d<typename vnd<SC,3>::template div_res_sc<Q>::type>
	(data/q);
    }


    template<typename Q>typename vnd<SC,3>::template mult_res_sc<Q>::type const
    dot(v3d<Q> const & v) const { return data.dot(v.data); }
    typename vnd<SC,3>::real_part const norm2() const { return data.norm2(); }
  };

  template<class SC> class v2d{
  private:
    vnd<SC,2> data;
    explicit v2d(vnd<SC,2> const &v_):data(v_) {}
  public:
    v2d() {}
    v2d(SC const & x,SC const & y) : data(x,y){}
    SC & operator[](unsigned int i) { return data[i]; }
    SC const &  operator[](unsigned int i) const { return data[i]; }
    std::ostream & print(std::ostream & C) const { return data.print(C); }
    std::istream & get(std::istream & C) { return data.get(C); }

    template<typename Q>
    typename vnd<SC,2>::template mult_res_sc<Q>::type const
    cross(v2d<Q> const & p){ return 
	typename vnd<SC,2>::template mult_res_sc<Q>::type(data.cross(p.data));
    }

    template<typename Q>
    v2d<SC> & operator*=(Q const & s){ data*= s; return *this; }
    template<typename Q>
    v2d<SC> & operator/=(Q const & s){ data/= s; return *this; }
    v2d<SC> operator+=(v2d<SC> const &v){
      data+=v.data; return *this;
    }
    v2d<SC> operator-=(v2d<SC> const &v){
      data-=v.data; return *this;
    }

    template<typename Q>
    v2d<typename vnd<SC,2>::template add_res_sc<Q>::type> const
    operator+(v2d<Q> const &v) const {
      return v2d<typename vnd<SC,2>::template add_res_sc<Q>::type>
	(data+v.data);
    }
    template<typename Q>
    v2d<typename vnd<SC,2>::template add_res_sc<Q>::type> const
    operator-(v2d<Q> const &v) const {
      return v2d<typename vnd<SC,2>::template add_res_sc<Q>::type>
	(data-v.data);
    }
    template<typename Q>
    v2d<typename vnd<SC,2>::template mult_res_sc<Q>::type> const
    operator*(Q const &q) const {
      return v2d<typename vnd<SC,2>::template mult_res_sc<Q>::type>
	(data*q);
    }
    template<typename Q>
    v2d<typename vnd<SC,2>::template div_res_sc<Q>::type> const
    operator/(Q const &q) const {
      return v2d<typename vnd<SC,2>::template div_res_sc<Q>::type>
	(data/q);
    }


    template<typename Q>typename vnd<SC,2>::template mult_res_sc<Q>::type const
    dot(v2d<Q> const & v) const { return data.dot(v.data); }
    typename vnd<SC,2>::real_part const norm2() const { return data.norm2(); }
  };

  template<typename SC,typename Q>
  v3d<typename mult_promotion_traits<SC,Q>::promote_type> const
  cross(v3d<SC> const & a,v3d<Q> const & b){ return a.cross(b); }
  template<typename SC,typename Q>
  typename mult_promotion_traits<SC,Q>::promote_type const
  cross(v2d<SC> const & a,v2d<Q> const & b){ return a.cross(b); }

  template<typename Q,typename SC>
  v3d<typename mult_promotion_traits<Q,SC>::promote_type> const
  operator*(Q const &q,v3d<SC> const & v)
  {return v*q;}
  template<typename Q,typename SC>
  v2d<typename mult_promotion_traits<Q,SC>::promote_type> const
  operator*(Q const &q,v2d<SC> const & v)
  {return v*q;}

  template<typename SC> v3d<SC> const conj(v3d<SC> v){ return v.conj(); }
  template<typename SC> v2d<SC> const conj(v2d<SC> v){ return v.conj(); }

  template<typename SC> typename vnd<SC,3>::real_part const norm2(v3d<SC> v)
  { return v.norm2(); }
  template<typename SC> typename vnd<SC,2>::real_part const norm2(v2d<SC> v)
  { return v.norm2(); }

  template<typename SC,typename Q>
  typename vnd<SC,N>::template mult_res_sc<Q>::type const
  dot(v3d<SC> const &v1,v3d<Q> const &v2){return v1.dot(v2); }
  template<typename SC,typename Q>
  typename vnd<SC,N>::template mult_res_sc<Q>::type const
  dot(v2d<SC> const &v1,v2d<Q> const &v2){return v1.dot(v2); }

  template<typename SC> std::ostream &
  operator<< (std::ostream & C,v3d<SC> const & a){ return a.print(C); }
  template<typename SC> std::ostream &
  operator<< (std::ostream & C,v2d<SC> const & a){ return a.print(C); }

  template<typename SC> std::istream & 
  operator >> (std::istream & C,v3d<SC> & a){ return a.get(C);}
  template<typename SC> std::istream & 
  operator >> (std::istream & C,v2d<SC> & a){ return a.get(C);}

} // of namespace pu

// Let's say such intervention does not seems so good.
// But I'm trying to work directly with boost ublas.
#include<boost/numeric/ublas/traits.hpp>
namespace boost {  namespace numeric { namespace ublas {
  // for multiplications
  template<typename Q,typename SC> struct promote_traits<Q,pu::v3d<SC> >{
    typedef pu::v3d< typename pu::vnd<SC,3>::template mult_res_sc<Q>::type >
    promote_type;
  };
  template<typename SC,typename Q> struct promote_traits<pu::v3d<SC>,Q>{
    typedef pu::v3d< typename pu::vnd<SC,3>::template mult_res_sc<Q>::type >
    promote_type;
  };

  // for additions/substractions
  template<typename SC,typename Q>
  struct promote_traits<pu::v3d<SC>,pu::v3d<Q> >{
    typedef pu::v3d< typename pu::vnd<SC,3>::template add_res_sc<Q>::type >
    promote_type;
  };
 }}
}

#endif
